package de.halbekunst.juplo.cachecontrol;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.io.PrintWriter;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.Collections;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.Locale;
import java.util.Map;
import java.util.zip.GZIPOutputStream;
import javax.servlet.Filter;
import javax.servlet.FilterChain;
import javax.servlet.FilterConfig;
import javax.servlet.ServletException;
import javax.servlet.ServletOutputStream;
import javax.servlet.ServletRequest;
import javax.servlet.ServletResponse;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.servlet.http.HttpServletResponseWrapper;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowire;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Configurable;



/**
 *
 * @author kai
 */
@Configurable(autowire=Autowire.BY_NAME)
public class AcceleratorFilter implements Filter {
  private final static Logger log = LoggerFactory.getLogger(AcceleratorFilter.class);

  private final static Map<String,String> EMPTY = Collections.unmodifiableMap(new HashMap<String,String>());

  public final static String REQUEST_URI_ATTRIBUTE = "javax.servlet.include.request_uri";
  public final static String RESPONSE_WRAPPER = AcceleratorFilter.class.getName() + ".RESPONSE_WRAPPER";


  @Autowired CacheControl cacheControl;
  @Autowired(required=true) Integer buffer;
  @Autowired(required=true) String eTag;
  @Autowired(required=true) Boolean weak;
  @Autowired(required=true) Long lastModified;
  @Autowired(required=true) Integer cacheSeconds;


  @Override
  public void doFilter(ServletRequest request, ServletResponse response, FilterChain chain) throws IOException, ServletException {
    if (!(request instanceof HttpServletRequest)) {
      log.error("AcceleratorFilter can only handle HTTP-requests");
      chain.doFilter(request, response);
      return;
    }

    /** Prüfen, ob es sich um eine Anfrage für einen JSP-Include handelt */
    if (request.getAttribute(REQUEST_URI_ATTRIBUTE) != null) {
      log.debug("Includes cannot be accelerated");
      chain.doFilter(request, response);
      return;
    }

    HttpServletRequest httpRequest = (HttpServletRequest)request;
    HttpServletResponse httpResponse = (HttpServletResponse)response;
    AccelerationWrapper wrapper = new AccelerationWrapper(httpRequest, httpResponse);
    httpRequest.setAttribute(RESPONSE_WRAPPER, wrapper);
    cacheControl.init(wrapper);
    chain.doFilter(request, wrapper);
    wrapper.finish();
  }

  @Override
  public void init(FilterConfig filterConfig) throws ServletException {
  }

  @Override
  public void destroy() {
  }


  class AccelerationWrapper extends HttpServletResponseWrapper implements CacheMethodHandle {

    private final HttpServletRequest request;
    private final HttpServletResponse response;

    boolean zipped; // CacheControll greift direkt auf dieses Flag zu!
    boolean forceCompression = false;

    private CountingServletOutputStream out;
    private ServletOutputStream stream;
    private PrintWriter writer;

    private boolean guessing = true;

    private long now;
    private int status;
    private int cacheSeconds;
    private boolean cacheSecondsSet = false;
    private long lastModified, expires = 0l;
    private String eTag;
    private boolean weak;
    private Map<String,String> cacheParams;


    AccelerationWrapper(HttpServletRequest request, HttpServletResponse response) throws IOException {
      super(response);

      this.request = request;
      this.response = response;

      now = System.currentTimeMillis();
      status = HttpServletResponse.SC_OK;
      cacheSeconds = AcceleratorFilter.this.cacheSeconds;
      lastModified = AcceleratorFilter.this.lastModified;
      eTag = AcceleratorFilter.this.eTag;
      weak = AcceleratorFilter.this.weak;
      cacheParams = new HashMap<String,String>();

      zipped = false;
      Enumeration values = request.getHeaders(Headers.HEADER_ACCEPT_ENCODING);
      while (values.hasMoreElements()) {
        String value = (String) values.nextElement();
        if (value.indexOf("gzip") != -1) {
          zipped = true;
          break;
        }
      }
      if (zipped)
        out = new GZIPServletOutputStream();
      else
        out = new CountingServletOutputStream();
    }

    public void finish() throws IOException {
      out.close();
    }


    @Override
    public void setStatus(int sc) {
      response.setStatus(sc);
      status = sc;
      try {
        cacheControl.decorate(request, response, this);
      }
      catch (Exception e) {
        log.error("Error while decorating response", e);
      }
    }

    @Override
    public void setStatus(int sc, String sm) {
      response.setStatus(sc,sm);
      status = sc;
      try {
        cacheControl.decorate(request, response, this);
      }
      catch (Exception e) {
        log.error("Error while decorating response", e);
      }
    }

    @Override
    public void addDateHeader(String name, long value) {

      if (!guessing) {
        super.addDateHeader(name, value);
        return;
      }

      if (Headers.HEADER_DATE.equalsIgnoreCase(name)) {
        now = value;
        calculateCacheSeconds();
        return;
      }

      if (Headers.HEADER_EXPIRES.equalsIgnoreCase(name)) {
        expires = value;
        calculateCacheSeconds();
        return;
      }

      if (Headers.HEADER_LAST_MODIFIED.equalsIgnoreCase(name)) {
        lastModified = value;
        return;
      }

      /** Unknown header: pass throug! */
      super.addDateHeader(name, value);
    }

    @Override
    public void addHeader(String name, String value) {

      if (!guessing) {
        super.addHeader(name, value);
        return;
      }

      if (value == null)
        return;
      analyzeHeader(name, value, false);
    }

    @Override
    public void addIntHeader(String name, int value) {

      if (!guessing) {
        super.addIntHeader(name, value);
        return;
      }

      analyzeHeader(name, Integer.toString(value), false);
    }

    @Override
    public void setDateHeader(String name, long value) {

      if (!guessing) {
        super.setDateHeader(name, value);
        return;
      }

      if (Headers.HEADER_DATE.equalsIgnoreCase(name)) {
        now = value;
        calculateCacheSeconds();
        return;
      }

      if (Headers.HEADER_EXPIRES.equalsIgnoreCase(name)) {
        expires = value;
        calculateCacheSeconds();
        return;
      }

      if (Headers.HEADER_LAST_MODIFIED.equalsIgnoreCase(name)) {
        lastModified = value;
        return;
      }

      /** Unknown header: pass throug! */
      super.setDateHeader(name, value);
    }

    @Override
    public void setHeader(String name, String value) {

      if (!guessing) {
        super.setHeader(name, value);
        return;
      }

      analyzeHeader(name, value, true);
    }

    @Override
    public void setIntHeader(String name, int value) {

      if (!guessing) {
        super.setIntHeader(name, value);
        return;
      }

      analyzeHeader(name, Integer.toString(value), true);
    }

    @Override
    public ServletOutputStream getOutputStream() throws IOException {

      if (writer != null)
        throw new IllegalStateException("ServletOutputStream and PrintWriter cannot be requested both!");

      if (stream == null) {
        stream = out;
      }

      return out;
    }

    @Override
    public PrintWriter getWriter() throws IOException {

      if (stream != null)
        throw new IllegalStateException("ServletOutputStream and PrintWriter cannot be requested both!");

      if (writer == null) {
        writer = new PrintWriter(out);
      }

      return writer;
    }

    @Override
    public void setContentLength(int len) {
      if (zipped)
        log.info("Supressing explicit content-length {} for request {}, because content will be zipped!", len, request.getRequestURI());
      else
        response.setContentLength(len);
    }

    @Override
    public void setBufferSize(int size) {

      out.setBuffer(size);
      response.setBufferSize(size);
    }

    @Override
    public void flushBuffer() throws IOException {

      forceCompression = true;
      cacheControl.decorate(request, response, this);
      response.flushBuffer();
    }

    @Override
    public void resetBuffer() {

      response.resetBuffer();
      try {
        if (zipped)
          out = new GZIPServletOutputStream();
        else
          out = new CountingServletOutputStream();
      }
      catch (IOException e) {
        throw new IllegalStateException(e);
      }
      stream = null;
      writer = null;
    }

    @Override
    public void reset() {

      response.reset();
      try {
        if (zipped)
          out = new GZIPServletOutputStream();
        else
          out = new CountingServletOutputStream();
      }
      catch (IOException e) {
        throw new IllegalStateException(e);
      }
      stream = null;
      writer = null;

      /** Cookies has been cleared! Reinitialize decorator... */
      forceCompression = false;
      cacheControl.init(this);
    }



    @Override
    public boolean isZipped() {
      return out.isZipped();
    }

    @Override
    public long getTimestamp() {
      return now;
    }

    @Override
    public int accepts(HttpServletRequest request) {
      return status;
    }

    @Override
    public int getCacheSeconds(HttpServletRequest request) {
      return cacheSeconds;
    }

    @Override
    public long getLastModified(HttpServletRequest request) {
      return lastModified;
    }

    @Override
    public String getETag(HttpServletRequest request) {
      return eTag;
    }

    @Override
    public boolean isETagWeak() {
      return weak;
    }

    @Override
    public void cacheControl(HttpServletRequest request, Map<String, String> cacheControlMap) {
      cacheControlMap.putAll(cacheParams);
    }

    @Override
    public Map<String,String> getAdditionalHeaders(HttpServletRequest request) {
      return EMPTY;
    }

    public void supressGuessing() {
      guessing = false;
    }


    private void analyzeHeader(String name, String value, boolean overwrite) {
      if (name == null)
        return;
      name = name.trim();

      if (name.equalsIgnoreCase(Headers.HEADER_DATE)) {
        if (value == null) {
          if (overwrite) {
            now = System.currentTimeMillis();
            cacheSeconds = AcceleratorFilter.this.cacheSeconds;
          }
          return;
        }
        try {
          SimpleDateFormat parser = new SimpleDateFormat(Headers.RFC_1123_DATE_FORMAT, Locale.US);
          now = parser.parse(value).getTime();
          calculateCacheSeconds();
        }
        catch (ParseException e) {
          log.warn("ignoring date for header \"Date\" in invalid format: {}", value);
        }
        return;
      }

      if (name.equalsIgnoreCase(Headers.HEADER_EXPIRES)) {
        if (value == null) {
          if (overwrite) {
            expires = 0;
            cacheSeconds = AcceleratorFilter.this.cacheSeconds;
          }
          return;
        }
        try {
          SimpleDateFormat parser = new SimpleDateFormat(Headers.RFC_1123_DATE_FORMAT, Locale.US);
          expires = parser.parse(value).getTime();
          calculateCacheSeconds();
        }
        catch (ParseException e) {
          log.warn("ignoring date for header \"Expires\" in invalid format: {}", value);
        }
        return;
      }

      if (name.equalsIgnoreCase(Headers.HEADER_LAST_MODIFIED)) {
        if (value == null) {
          if (overwrite)
            lastModified = AcceleratorFilter.this.lastModified;
          return;
        }
        try {
          SimpleDateFormat parser = new SimpleDateFormat(Headers.RFC_1123_DATE_FORMAT, Locale.US);
          lastModified = parser.parse(value).getTime();
        }
        catch (ParseException e) {
          log.warn("ignoring date for header \"Last-Modified\" in invalid format: {}", value);
        }
        return;
      }

      if (name.equalsIgnoreCase(Headers.HEADER_ETAG)) {
        if (value == null) {
          if (overwrite) {
            eTag = AcceleratorFilter.this.eTag;
            weak = AcceleratorFilter.this.weak;
          }
          return;
        }
        value = value.trim();
        int start = 0;
        int end = value.length();
        if (value.startsWith("W/")) {
          weak = true;
          start = 2;
        }
        else {
          weak = false;
        }
        if (value.charAt(start) == '"')
          start++;
        else
          log.warn("Quote at the beginning ov ETag is missing: {}", value);
        if (value.charAt(end -1) == '"')
          end--;
        else
          log.warn("Quote at the end of ETag is missing: {}", value);
        eTag = value.substring(start, end);
        String filtered = eTag.replaceAll("[^\\x00-\\x21\\x23-\\x7F]+","");
        if (filtered.length() < eTag.length()) {
          log.warn("filtering out illegal characters in ETag: \"{}\" -> \"{}\"", eTag, filtered);
          eTag = filtered;
        }
      }

      if (name.equalsIgnoreCase(Headers.HEADER_CACHE_CONTROL)) {
        if (overwrite)
          cacheParams.clear();
        if (value == null)
          return;
        for (String param : value.split(",")) {
          param = param.trim();
          int pos = param.indexOf("=");
          if (pos < 0) {
            cacheParams.put(param, null);
          }
          else {
            String paramName = param.substring(0, pos).trim();
            if (paramName.equalsIgnoreCase("max-age")) {
              try {
                cacheSeconds = Integer.parseInt(param.substring(pos + 1));
                cacheSecondsSet = true;
              }
              catch (NumberFormatException e) {
                log.warn("illegal value for Header \"Cache-Control\":", param);
              }
            }
            else {
              cacheParams.put(paramName, param.substring(pos + 1));
            }
          }
        }
        return;
      }

      if (name.equalsIgnoreCase(Headers.HEADER_PRAGMA)) {
        if (value != null && value.trim().equalsIgnoreCase("no-cache"))
          cacheSeconds = 0;
        return;
      }

      /** Pass header through, if no value from intrest was found */
      if (overwrite)
        super.setHeader(name, value);
      else
        super.addHeader(name, value);
    }

    private void calculateCacheSeconds() {
      if (!cacheSecondsSet && expires >= now) {
        cacheSeconds = (int)(expires/1000 - now/1000);
        log.debug("calculating cache-seconds from DATE and EXPIRES: {}", cacheSeconds);
      }
    }


    class CountingServletOutputStream extends ServletOutputStream {

      private OutputStream out;
      int left;
      boolean empty;


      CountingServletOutputStream() throws IOException {
        out = response.getOutputStream();
        left = buffer;
        empty = true;
      }


      void setBuffer(int size) throws IllegalStateException {}

      boolean isZipped() {
        return false;
      }

      void finish() throws IOException {
        decorate();
      }

      void decorate() throws IOException {
        try {
          AcceleratorFilter.this.cacheControl.decorate(AccelerationWrapper.this.request, response, AccelerationWrapper.this);
        }
        catch (Exception e) {
          log.error("Error while guessing Cache-Header's", e);
          response.sendError(HttpServletResponse.SC_INTERNAL_SERVER_ERROR);
        }
      }


      @Override
      public void close() throws IOException {
        decorate();
        out.close();
      }

      @Override
      public void flush() throws IOException {
        decorate();
        out.flush();
      }

      @Override
      public void write(int i) throws IOException {
        empty = false;
        if (left == 0)
          decorate();
        left--;
        out.write(i);
      }
    }


    final class GZIPServletOutputStream extends CountingServletOutputStream {

      final static int MINIMAL_BUFFER_SIZE = 128;

      private ByteArrayOutputStream buffer;
      private OutputStream out;
      private GZIPOutputStream zout;
      private int bufferSize;
      private boolean decorated = false;


      public GZIPServletOutputStream() throws IOException {
        setBuffer(AcceleratorFilter.this.buffer);
      }


      @Override
      void finish() throws IOException {
        decorate();
        zout.finish();
        super.out.write(buffer.toByteArray());
      }

      @Override
      boolean isZipped() {
        if (decorated)
          return true;
        return !empty;
      }

      @Override
      void setBuffer(int size) throws IllegalStateException {
        if (!empty)
          throw new IllegalStateException("attemp to change buffer size after writing data to response!");

        if (size > MINIMAL_BUFFER_SIZE) {
          left = size;
          try {
            bufferSize = size;
            out = response.getOutputStream();
            buffer = new ByteArrayOutputStream(size);
            zout = new GZIPOutputStream(buffer, size);
            super.out = zout;
          }
          catch (IOException e) {
            throw new IllegalStateException(e);
          }
        }
      }



      @Override
      public void close() throws IOException {
        decorate();
        zout.close();
        out.write(buffer.toByteArray());
        out.close();
      }

      @Override
      public void flush() throws IOException {
        decorate();
        out.write(buffer.toByteArray());
        out.flush();
      }

      @Override
      public void write(int i) throws IOException {
        empty = false;
        if (left == 0) {
          if (!decorated) {
            decorate();
            decorated = true;
          }
          out.write(buffer.toByteArray());
          buffer.reset();
          left = bufferSize;
        }
        left--;
        zout.write(i);
      }
    }
  }
}
